/* doublehash.cc */
/* von Timo Bingmann */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>

#define word       unsigned int

// Prime Number Size Hash for Modulo Hash 
// const int HASHSIZE = 36701;
// const int HASHSIZE = 256*256;

#if defined(HASH_LINEAR_MODULO_TINY)
#define TESTNAME      "A tiny modulo hash with linear probing. Good for demonstration purposes."
#define HASHSIZE      37
#define FILLSIZE      18
#define DEBUG         1
#define PRINT_HASH
#define HASHFUNC      hashLinearModulo

#elif defined(HASH_LINEAR_MULTIPLY_MEDIUM)
#define TESTNAME      "A medium-size multiplication hash with linear probing. Still good for demonstration purposes."
#define HASHSIZE      4096
#define HASHBITS      12
#define FILLSIZE      2000
#define DEBUG         1
#define PRINT_HASH
#define HASHFUNC      hashLinearMultiply

#elif defined(HASH_LINEAR_MODULO)
#define TESTNAME      "A fairly large modulo hash with linear probing"
#define HASHSIZE      36701
#define FILLSIZE      18000
#define DEBUG         0
#define HASHFUNC      hashLinearModulo

#elif defined(HASH_SQUARE_MODULO) 
#define TESTNAME      "A modulo hash with square probing"
#define HASHSIZE      36701
#define FILLSIZE      18000
#define DEBUG         0
#define HASHFUNC      hashSquareModulo

#elif defined(HASH_SQUARE_MULTIPLY) 
#define TESTNAME      "A mulitplication-based hash (size = power of 2) with square probing"
#define HASHSIZE      65536
#define HASHBITS      16             // = log2(HASHSIZE)
#define FILLSIZE      32000
#define DEBUG         0
#define HASHFUNC      hashSquareMultiply

#elif defined(HASH_DOUBLE_MODULO) 
#define TESTNAME      "A double hash function with two modulo funcs"
#define HASHSIZE      36701
#define FILLSIZE      18000
#define DEBUG         0
#define HASHFUNC      hashDoubleModulo

#elif defined(HASH_DOUBLE_UNIVERSAL) 
#define TESTNAME      "A double hash function with two random universal hash funcs"
#define HASHSIZE      65536
#define HASHBITS      16             // = log2(HASHSIZE)
#define FILLSIZE      32000
#define DEBUG         0
#define HASHFUNC      hashDoubleUniversal

#elif defined(HASH_SQUARE_MULTIPLY_HUGE) 
#define TESTNAME      "A multiplication-based hash on a huge (16MB) hash table with square probing"
#define HASHSIZE      16777216
#define HASHBITS      24             // = log2(HASHSIZE)
#define FILLSIZE      8380000
#define DEBUG         0
#define HASHFUNC      hashSquareMultiply

#elif defined(HASH_DOUBLE_UNIVERSAL_HUGE)
#define TESTNAME      "Double-hashing function on a huge (16MB) hash table with two random universal hash funcs"
#define HASHSIZE      16777216
#define FILLSIZE      8380000
#define HASHFUNC      hashDoubleUniversal
#define DEBUG         0

#else

#error Youd better specify which hash example you want. Or create one here.

#endif

/* Large prime number hash table size
#define HASHSIZE      11983729
#define FILLSIZE      6000000
*/

// Two (fairly) good factors for multiplication hashing
const word MULTIFACTOR = 0x9E377969;   // = 2654435769 = (sqrt(5) - 1) / 2 * 2^32
const word MULTIFACTOR2 = 0x5db3d745;  // = 1572067139 = (sqrt(3) - 1) / 2 * 2^32

// the big thing.
void* hashArray[HASHSIZE];

// Prime Numbers for our two universal hashing functions
// Rand[AB][12] get filled by rand in main
// int UniversalHashPrime1 = 69259;
int UniversalHashPrime1 = 19126033;
int RandA1, RandB1;

// int UniversalHashPrime2 = 69029;
int UniversalHashPrime2 = 20803777;
int RandA2, RandB2;

// HASHBITS is only used by hashMultiply
#if !defined(HASHBITS)
#define HASHBITS 8
#endif

// ***** Debug Messages ******

#ifndef DEBUG
#define DEBUG      0
#warning Define a Debuglevel with -DDEBUG=#
#warning Level 0: Just statical output
#warning Level 1: Each key/hash operation printed
#warning Level 2: Full hash output after each operation
#endif

// depending on DEBUG these three macros either expand to nullprint or printf

#if DEBUG >= 0
#define dbg0      printf
#else
#define dbg0      nullprint
#endif

#if DEBUG >= 1
#define dbg1      printf
#else
#define dbg1      nullprint
#endif

#if DEBUG >= 2
#define dbg2      printf
#else
#define dbg2      nullprint
#endif

int nullprint(const char *format, ...) {
  return 0;
}

// ***** Hash Function ******

// Primitive hash functions
word hashModulo(void *key) {
  return ((word)key) % HASHSIZE;
}
word hashMultiply(void *key) {
  return ((word)key * MULTIFACTOR) >> (32 - (HASHBITS));
}
word hashUniversal(void *key,word a, word b, word p) {
  return (((word)key * a + b) % p) % HASHSIZE;
}
word hashModulo2(void *key) {
  return 1 + ((word)key % (HASHSIZE-3));
}

// open-addressing with linear probing
word hashLinearModulo(void *key,int index) {
  return (hashModulo(key) + index) % HASHSIZE;
}
word hashLinearMultiply(void *key,int index) {
  return (hashMultiply(key) + index) % HASHSIZE;
}

// square probing
word hashSquareModulo(void *key,int index) {
  return (hashModulo(key) + index*index) % HASHSIZE;
}
word hashSquareMultiply(void *key,int index) {
  return (hashMultiply(key) + index*index) % HASHSIZE;
}

// double hashes
word hashDoubleModulo(void *key,int index) {
  return (hashModulo(key) + index*hashModulo2(key)) % HASHSIZE;
}
word hashDoubleMultiply(void *key,int index) {
  return (hashMultiply(key) + index*hashUniversal(key,RandA1,RandB2,UniversalHashPrime1)) % HASHSIZE;
}
word hashDoubleUniversal(void *key, int index) {
  return (hashUniversal(key,RandA1,RandB1,UniversalHashPrime1) + 
	  index * hashUniversal(key,RandA2,RandB2,UniversalHashPrime2)) % HASHSIZE;
}

// this function pointer is the one used for hashing. HASHFUNC macro is declared in the header

word (*hashFunction)(void *key, int index) = &HASHFUNC;

// ***** Hash Array Functions ******

void initializeHash() {
  memset(hashArray,0,sizeof hashArray);
}

bool insertHash(void *key,int *probes) {
  int index = 0;
  word hashVal = 0;

  while (index < HASHSIZE && 
	 hashArray[hashVal = hashFunction(key,index)] != NULL) {
    dbg2("J");
    index++;
  }
  *probes = index+1; // This is the number we're interested in: The number of collisions.

  if (index < HASHSIZE) {
    dbg1("\t hash %d ",hashVal);
    hashArray[hashVal] = key;
    return true;
  }
  else {
    dbg0("\nHash overflow. That's all folks: Press start to retry.\n");
    return false;
  }
}

void* findHash(void *key,int *probes) {
  int index = 0;
  word hashVal = 0;

  // Arguments are evaluated from left to right, right?
  while(index < HASHSIZE &&
	hashArray[hashVal = hashFunction(key,index)] != key &&
	hashArray[hashVal] != NULL) { 
    index++;
  }
  *probes = index+1; // This is the number we're interested in: The number of collisions.

  if (index < HASHSIZE && hashArray[hashVal] == key) 
    return key;
  else
    return NULL;
}

// primitive hash output function: print every filled cell as * between >   <
void printHash() {
  dbg1("\tArray: >");
  for(int i=0;i<HASHSIZE;i++) {
    char c = hashArray[i] != NULL ? '*' /* ((int)hashArray[i] % 26) + 'a'*/ : ' ';
    dbg1("%c",c);
  }
  dbg1("<");
}

// bit more advanced output function: print the number of filled cells
// in 10 subsequent cell groups as 0-9* between >   < 
void printHashGrouped() {
  dbg1("\tArray: >");
  //  const char symbols[SYMS+1] = " 12345678*";
  const char symbols[] = " 123456789*";
  const int SYMS = 10;
  for(int i = 0;i < HASHSIZE;i += SYMS) {
    int index = 0;
    for(int a = 0;a < SYMS && i+a < HASHSIZE;a++)
      index += hashArray[i+a] != NULL ? 1 : 0;
    dbg1("%c",symbols[index]);
  }
  dbg1("<\n");
}

// Small functions to handle average counting in an intuative way.
struct AvgCounter {
  int total;
  int num;
};

void AvgAdd(AvgCounter *cnt, int num) {
  cnt->total += num;
  cnt->num++;
}

double AvgGet(AvgCounter *cnt) {
  return (double)cnt->total / (double)cnt->num;
}

int main() {

  initializeHash();

#define XSTR(x) STR(x)
#define STR(x) #x

  dbg0("Running hashing tests with " XSTR(HASHFUNC) "\n");
  dbg0("Hash Test: " TESTNAME "\n");

  srand(time(0));
  RandA1 = 1234 + ( rand() % (UniversalHashPrime1 - 1234) );
  RandB1 = 2345 + ( rand() % (UniversalHashPrime1 - 2345) );  
  RandA2 = 1234 + ( rand() % (UniversalHashPrime2 - 1234) );
  RandB2 = 2345 + ( rand() % (UniversalHashPrime2 - 2345) );

  // RandA1 = 26801;
  // RandB1 = 23743;
  // RandA2 = 9719;
  // RandB2 = 41681;

  if (hashFunction == hashDoubleUniversal) {
    dbg0("Random Numbers for 1st Univsal Hashing Function: A = %d; B = %d; P = %d\n",
       RandA1,RandB1,UniversalHashPrime1);
    dbg0("Random Numbers for 2nd Univsal Hashing Function: A = %d; B = %d; P = %d\n",
       RandA2,RandB2,UniversalHashPrime2);
  }

  // Counting Variables
  AvgCounter Inserts;
  Inserts.total = Inserts.num = 0;
  int fill = FILLSIZE;

  double alpha = double(fill) / double(HASHSIZE);

  dbg0("Filling our hash (HASHSIZE = %d) with %d items. Hash load will be %.2f\n",
       HASHSIZE,fill,alpha);
  dbg0("Our Uniform Hashing analysis lets us expect:\n");
  dbg0("\t %.2f probes for an unsuccessful search: 1/(1-alpha)\n",1.0/(1.0-alpha));
  dbg0("\t %.2f probes for a successful search: 1/alpha * ln(1/(1-alpha))\n",
       1.0/alpha * log(1.0/(1.0-alpha)));

#if defined(PRINT_HASH)
  dbg0("=== Press return to continue.\n");
  getchar();
#endif

  for(int m=0;m<fill;m++) {

    int x = (rand() % fill) + 1 ;
    // x = ((m) % fill) + 1;

    dbg1("%5d> Inserting %11d:",m,x);

    int probes;
    if (! insertHash((void*)x,&probes)) {
      exit(-1);
    }

    AvgAdd(&Inserts,probes);

    dbg2("\tProbes: %2d Avg: %.2f\n",probes,AvgGet(&Inserts));

#if defined(PRINT_HASH)
    printHash();
#endif

    // printHashGrouped();
    dbg1("\n");
  }

  // printHashGrouped();
  dbg0("==========================================\n");
  dbg0(" Insertion avged to %.2f probes/insert\n",AvgGet(&Inserts));
  dbg0(" %d total keys inserted. %d total probes taken.\n", Inserts.num, Inserts.total);
  dbg0(" Finished inserting. Let's now look for a few keys.\n");

#if defined(PRINT_HASH)
  dbg0("=== Press return to continue.\n");
  getchar();
#endif

  AvgCounter Searches, SuccessSearches, FailedSearches;

  Searches.num = Searches.total = 0;
  SuccessSearches.num = SuccessSearches.total = 0;
  FailedSearches.num = FailedSearches.total = 0;

  int keyspace = HASHSIZE - 1;
  // keyspace = fill;

  for(int n=0;n<keyspace;n++) {
    
    int x = (rand() % keyspace) + 1;
    // x = (n % keyspace) + 1;

    dbg1("%3d> finding %6d ",n,x);

    int probes;
    if (! findHash((void*)x,&probes)) {
      dbg1("\tnope not found. ");
      AvgAdd(&FailedSearches,probes);
    }
    else {
      dbg1("\tneedle found. Ouch! ");
      AvgAdd(&SuccessSearches,probes);
    }

    AvgAdd(&Searches,probes);

    dbg1("\t%2d probes avg: %.2f. total %7d / %6d",probes,
	 AvgGet(&Searches),Searches.total,Searches.num);

    dbg1("\n");
  }

  dbg0("==========================================\n");
  dbg0(" Finding avged to %.2f probes/search\n",AvgGet(&Searches));
  dbg0(" %d total keys looked-up. %d total probes takes.\n",Searches.num,Searches.total);
  dbg0(" of these %d lookups where successful with %d probes.\n",
       SuccessSearches.num,SuccessSearches.total);
  dbg0(" with averages to %.2f probes/successful search,\n",AvgGet(&SuccessSearches));
  dbg0(" and %d lookups which failed with %d failing probes.\n",
       FailedSearches.num,FailedSearches.total);
  dbg0(" averaging to %.2f probes/unsuccessful search.\n",AvgGet(&FailedSearches));
  
  return 0;
}
